# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import StrEnum
from typing import Optional

from pydantic import BaseModel


class UiType(StrEnum):
    chat = "chat"
    hands_off = "hands-off"
    custom = "custom"


class CliExample(BaseModel, extra="allow"):
    command: str
    name: Optional[str] = None
    description: Optional[str] = None
    output: Optional[str] = None
    processingSteps: Optional[list[str]] = None


class UiDefinition(BaseModel, extra="allow"):
    type: UiType
    userGreeting: str | None = None


class Examples(BaseModel, extra="allow"):
    cli: Optional[list[CliExample]] = None


class Metadata(BaseModel, extra="allow"):
    title: Optional[str] = None
    fullDescription: Optional[str] = None
    framework: Optional[str] = None
    license: Optional[str] = None
    languages: Optional[list[str]] = None
    githubUrl: Optional[str] = None
    examples: Optional[Examples] = None
    avgRunTimeSeconds: Optional[float] = None
    avgRunTokens: Optional[float] = None
    tags: Optional[list[str]] = None
    ui: Optional[UiDefinition] = None
    provider: Optional[str] = None



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated, Literal, Union
from pydantic import BaseModel, Discriminator

from beeai_sdk.schemas.base import Input, Output


class UserMessage(BaseModel):
    role: Literal["user"]
    content: str


class AssistantMessage(BaseModel):
    role: Literal["assistant"]
    content: str


Message = Annotated[Union[UserMessage, AssistantMessage], Discriminator("role")]


class MessageInput(Input):
    messages: list[Message]


class MessageOutput(Output):
    messages: list[Message]



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
from contextlib import asynccontextmanager, AsyncExitStack
from typing import AsyncGenerator

import anyio

from acp import ClientSession, types, ServerNotification
from acp.client.sse import sse_client
from acp.shared.session import ReceiveResultT
from acp.types import RequestParams

logger = logging.getLogger(__name__)


@asynccontextmanager
async def mcp_client(url: str) -> AsyncGenerator[ClientSession, None]:
    """Context manager for MCP client connection."""
    async with sse_client(url=url) as (read_stream, write_stream):
        async with ClientSession(read_stream, write_stream) as session:
            await session.initialize()
            yield session


async def send_request_with_notifications(
    url_or_session: str | ClientSession, req: types.Request, result_type: type[ReceiveResultT]
) -> AsyncGenerator[ReceiveResultT | ServerNotification | None, None]:
    resp: ReceiveResultT | None = None
    async with AsyncExitStack() as exit_stack:
        if isinstance(url_or_session, str):
            session = await exit_stack.enter_async_context(mcp_client(url=url_or_session))
            await session.initialize()
        else:
            session = url_or_session

        message_writer, message_reader = anyio.create_memory_object_stream()

        req = types.ClientRequest(req).root
        req.params = req.params or RequestParams()
        req.params.meta = RequestParams.Meta(progressToken=uuid.uuid4().hex)
        req = types.ClientRequest(req)

        async with anyio.create_task_group() as task_group:

            async def request_task():
                nonlocal resp
                try:
                    resp = await session.send_request(req, result_type)
                finally:
                    task_group.cancel_scope.cancel()

            async def read_notifications():
                # IMPORTANT(!) if the client does not read the notifications, it gets blocked never receiving the response
                async for message in session.incoming_messages:
                    try:
                        if isinstance(message, Exception):
                            raise message
                        notification = ServerNotification.model_validate(message)
                        await message_writer.send(notification)
                    except ValueError:
                        logger.warning(f"Unable to parse message from server: {message}")

            task_group.start_soon(read_notifications)
            task_group.start_soon(request_task)

            async for message in message_reader:
                yield message
    if resp:
        yield resp


async def send_request(url_or_session: str, req: types.Request, result_type: type[ReceiveResultT]) -> ReceiveResultT:
    async for message in send_request_with_notifications(url_or_session, req, result_type):
        if isinstance(message, result_type):
            return message
    raise RuntimeError(f"No response of type {result_type.__name__} was returned")



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource, SERVICE_NAME, SERVICE_NAMESPACE
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExportResult
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from acp.server.highlevel import Server


class SilentOTLPSpanExporter(OTLPSpanExporter):
    def export(self, spans):
        try:
            return super().export(spans)
        except Exception as e:
            logging.debug(f"OpenTelemetry Exporter failed silently: {e}")
            return SpanExportResult.FAILURE


async def run_agent_provider(server: Server):
    server.settings.host = os.getenv("HOST", "127.0.0.1")
    server.settings.port = int(os.getenv("PORT", "8000"))
    trace.set_tracer_provider(
        tracer_provider=TracerProvider(
            resource=Resource(attributes={SERVICE_NAME: server.name, SERVICE_NAMESPACE: "beeai-agent-provider"}),
            active_span_processor=BatchSpanProcessor(SilentOTLPSpanExporter()),
        )
    )
    with trace.get_tracer("beeai-sdk").start_as_current_span("agent-provider"):
        try:
            await server.run_sse_async(timeout_graceful_shutdown=5)
        except KeyboardInterrupt:
            pass



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import StrEnum

from pydantic import BaseModel, Field


class Config(BaseModel, extra="allow"):
    pass


class Input(BaseModel, extra="allow"):
    config: Config | None = None


class LogLevel(StrEnum):
    error = "error"
    warning = "warning"
    info = "info"
    cite = "cite"
    success = "success"


class Log(BaseModel, extra="allow"):
    level: LogLevel = LogLevel.info
    message: str


class Output(BaseModel, extra="allow"):
    logs: list[Log | None] = Field(default_factory=list)
